@@ -1,25 +1,25 @@
 #!/usr/bin/env python3
 
 import asyncio
-import io
 import logging
+import math
 import os
 import socket
 import time
 from collections import defaultdict
 from concurrent.futures import Executor, ThreadPoolExecutor
-from typing import cast, List, Optional, Set
+from typing import cast, ClassVar, List, Optional, Set
 
 import psutil
 
-from .io_types import BufferType, IOReq, ReadReq, StoragePlugin, WriteReq
+from .io_types import BufferType, ReadIO, ReadReq, StoragePlugin, WriteIO, WriteReq
 from .pg_wrapper import PGWrapper
 
 logger: logging.Logger = logging.getLogger(__name__)
 
 
 _MAX_PER_RANK_MEMORY_BUDGET_BYTES: int = 32 * 1024 * 1024 * 1024
-_AVAILABLE_MEMORY_MULTIPLIER: float = 0.8
+_AVAILABLE_MEMORY_MULTIPLIER: float = 0.6
 _MAX_PER_RANK_CPU_CONCURRENCY: int = 4
 _MAX_PER_RANK_IO_CONCURRENCY: int = 16
 
@@ -69,33 +69,154 @@ class _WritePipeline:
 
     async def stage_buffer(self, executor: Executor) -> "_WritePipeline":
         self.buf = await self.write_req.buffer_stager.stage_buffer(executor)
-        if isinstance(self.buf, bytes):
-            self.buf_sz_bytes = len(self.buf)
-        else:
-            # self.buf is memoryview
-            self.buf_sz_bytes = self.buf.nbytes
+        self.buf_sz_bytes = len(self.buf)
         return self
 
     async def write_buffer(self) -> "_WritePipeline":
         if self.buf is None:
             raise AssertionError("self.buf can not be None.")
-        # pyre-ignore[6]: it's valid to initialize BytesIO with memoryview
-        # according to: https://docs.python.org/3/library/io.html#io.BytesIO
-        io_req = IOReq(path=self.write_req.path, buf=io.BytesIO(self.buf))
-        await self.storage.write(io_req)
+        write_io = WriteIO(path=self.write_req.path, buf=self.buf)
+        await self.storage.write(write_io=write_io)
 
         # Reclaim buffer memory
-        del io_req
+        del write_io
         self.buf = None
         return self
 
 
+_LOG_LINE_LIMIT = 8
+
+
+class _WriteReporter:
+    _WRITE_LOG_TEMPLATE: ClassVar[str] = (
+        "{rank:>4} {stageable:>12} {staging:>12} {writable:>12} "
+        "{writing:>12} {rss_delta:>16} {memory_budget:>20} {data_written:>18}"
+    )
+
+    def __init__(
+        self,
+        ready_for_staging: Set[_WritePipeline],
+        staging_tasks: Set[asyncio.Task],
+        ready_for_io: Set[_WritePipeline],
+        io_tasks: Set[asyncio.Task],
+        rank: int,
+        total_memory_budget_bytes: int,
+    ) -> None:
+        self.ready_for_staging = ready_for_staging
+        self.staging_tasks = staging_tasks
+        self.ready_for_io = ready_for_io
+        self.io_tasks = io_tasks
+        self.rank = rank
+        self.total_memory_budget_bytes = total_memory_budget_bytes
+
+        self._process = psutil.Process()
+        self.baseline_rss_bytes: int = self._process.memory_info().rss
+        self.begin_ts: float = time.monotonic()
+        self.bytes_written = 0
+
+        self._header: str = self._WRITE_LOG_TEMPLATE.format(
+            rank="Rank",
+            stageable="Stage-able",
+            staging="Staging",
+            writable="Writable",
+            writing="Writing",
+            rss_delta="RSS Delta (GB)",
+            memory_budget="Memory Budget (GB)",
+            data_written="Data Written (GB)",
+        )
+
+    def print_header(self) -> None:
+        if self.rank == 0:
+            logger.info(self._header)
+            logger.info("-" * len(self._header))
+
+    def report(self, memory_budget_bytes: int) -> None:
+        rss_bytes = self._process.memory_info().rss
+        rss_delta_gb = (rss_bytes - self.baseline_rss_bytes) / 1024**3
+        msg = self._WRITE_LOG_TEMPLATE.format(
+            rank=self.rank,
+            stageable=len(self.ready_for_staging),
+            staging=len(self.staging_tasks),
+            writable=len(self.ready_for_io),
+            writing=len(self.io_tasks),
+            rss_delta=f"{rss_delta_gb:.2f}",
+            memory_budget=(
+                f"{memory_budget_bytes / 1024**3:.2f}/"
+                f"{self.total_memory_budget_bytes / 1024**3:.2f}"
+            ),
+            data_written=f"{self.bytes_written / 1024**3:.2f}",
+        )
+        logger.info(msg)
+
+    def report_staging_done(self) -> None:
+        msg = (
+            f"Rank {self.rank} completed staging in "
+            f"{time.monotonic() - self.begin_ts:.2f} seconds"
+        )
+        logger.info(self._pad_msg(msg=msg))
+
+    def report_writing_done(self) -> None:
+        elapsed_secs = time.monotonic() - self.begin_ts
+        mbps = self.bytes_written / 1024**2 / (elapsed_secs)
+        msg = (
+            f"Rank {self.rank} completed writing in "
+            f"{elapsed_secs:.2f} seconds (throughput {mbps:.2f}MB/s)"
+        )
+        logger.info(self._pad_msg(msg=msg))
+
+    def _pad_msg(self, msg: str) -> str:
+        padding = (len(self._header) - len(msg) - 2) / 2
+        return f"{'-' * math.ceil(padding)} {msg} {'-' * math.floor(padding)}"
+
+
+class PendingIOWork:
+    def __init__(
+        self,
+        ready_for_io: Set[_WritePipeline],
+        io_tasks: Set[asyncio.Task],
+        memory_budget_bytes: int,
+        write_reporter: _WriteReporter,
+    ) -> None:
+        self.ready_for_io = ready_for_io
+        self.io_tasks = io_tasks
+        self.memory_budget_bytes = memory_budget_bytes
+        self.write_reporter = write_reporter
+        self.logging_freq: int = math.ceil(
+            (len(self.ready_for_io) + len(self.io_tasks)) / _LOG_LINE_LIMIT
+        )
+
+    async def complete(self) -> None:
+        while len(self.ready_for_io) + len(self.io_tasks) != 0:
+            done, _ = await asyncio.wait(
+                self.io_tasks, return_when=asyncio.FIRST_COMPLETED
+            )
+            for d in done:
+                write_pipeline: _WritePipeline = d.result()
+                self.memory_budget_bytes += cast(int, write_pipeline.buf_sz_bytes)
+                self.write_reporter.bytes_written += cast(
+                    int, write_pipeline.buf_sz_bytes
+                )
+                self.io_tasks.remove(d)
+                for p in set(self.ready_for_io):
+                    if len(self.io_tasks) >= _MAX_PER_RANK_IO_CONCURRENCY:
+                        break
+                    io_task = asyncio.create_task(p.write_buffer())
+                    self.io_tasks.add(io_task)
+                    self.ready_for_io.remove(p)
+            if (len(self.ready_for_io) + len(self.io_tasks)) % self.logging_freq == 0:
+                self.write_reporter.report(memory_budget_bytes=self.memory_budget_bytes)
+        self.write_reporter.report_writing_done()
+
+    def sync_complete(self, event_loop: asyncio.AbstractEventLoop) -> None:
+        event_loop.run_until_complete(self.complete())
+
+
 async def execute_write_reqs(
     write_reqs: List[WriteReq],
     storage: StoragePlugin,
     memory_budget_bytes: int,
     rank: int,
-) -> None:
+) -> PendingIOWork:
     # This function fulfills the write requests by moving them through the
     # following stages with the specified memory budget:
     #
@@ -105,17 +226,25 @@ async def execute_write_reqs(
     # io - Writing the buffer to the storage.
     # done (implicit) - The write request has been fulfilled.
     #
-    # TODO: for async checkpointing, as soon as all write requests move past
-    # the staging stage, this function can return and continue writing to
-    # storage in the background.
+    # It returns as soon as all write requests are moved past the staging
+    # stage.
     ready_for_staging = {_WritePipeline(write_req, storage) for write_req in write_reqs}
     staging_tasks = set()
-    ready_for_io = set()
-    io_tasks = set()
-    executor = ThreadPoolExecutor(max_workers=_MAX_PER_RANK_CPU_CONCURRENCY)
+    ready_for_io: Set[_WritePipeline] = set()
+    io_tasks: Set[asyncio.Task] = set()
 
-    bytes_written = 0
-    begin_ts = time.monotonic()
+    write_reporter = _WriteReporter(
+        ready_for_staging=ready_for_staging,
+        staging_tasks=staging_tasks,
+        ready_for_io=ready_for_io,
+        io_tasks=io_tasks,
+        rank=rank,
+        total_memory_budget_bytes=memory_budget_bytes,
+    )
+    logging_freq = math.ceil(len(ready_for_staging) / _LOG_LINE_LIMIT)
+    write_reporter.print_header()
+
+    executor = ThreadPoolExecutor(max_workers=_MAX_PER_RANK_CPU_CONCURRENCY)
 
     def dispatch_staging(
         ready_for_staging: Set[_WritePipeline],
@@ -127,7 +256,12 @@ async def execute_write_reqs(
         Dispatch as many staging tasks as the memory budget allows.
         """
         for p in set(ready_for_staging):
-            if len(staging_tasks) == 0 or p.staging_cost_bytes < memory_budget_bytes:
+            # Allow staging tasks whose cost exceeds the memory budget only
+            # when there is no inflight write pipeline.
+            if (
+                len(staging_tasks) + len(ready_for_io) + len(io_tasks) == 0
+                or p.staging_cost_bytes < memory_budget_bytes
+            ):
                 memory_budget_bytes -= p.staging_cost_bytes
                 staging_task = asyncio.create_task(p.stage_buffer(executor))
                 staging_tasks.add(staging_task)
@@ -154,10 +288,7 @@ async def execute_write_reqs(
         executor=executor,
     )
 
-    while (
-        len(ready_for_staging) + len(staging_tasks) + len(ready_for_io) + len(io_tasks)
-        != 0
-    ):
+    while len(ready_for_staging) + len(staging_tasks) != 0:
         done, _ = await asyncio.wait(
             staging_tasks | io_tasks, return_when=asyncio.FIRST_COMPLETED
         )
@@ -171,17 +302,14 @@ async def execute_write_reqs(
                 # torch.save, the staging cost is 2x the buffer size.
                 memory_budget_bytes += write_pipeline.staging_cost_bytes
                 memory_budget_bytes -= cast(int, write_pipeline.buf_sz_bytes)
+                if len(staging_tasks) % logging_freq == 0:
+                    write_reporter.report(memory_budget_bytes=memory_budget_bytes)
 
-                if len(ready_for_staging) == 0 and len(staging_tasks) == 0:
-                    logger.debug(
-                        f"Rank {rank} finished staging in "
-                        f"{time.monotonic() - begin_ts:.2f} seconds"
-                    )
             elif d in io_tasks:
                 io_tasks.remove(d)
                 write_pipeline: _WritePipeline = d.result()
                 memory_budget_bytes += cast(int, write_pipeline.buf_sz_bytes)
-                bytes_written += cast(int, write_pipeline.buf_sz_bytes)
+                write_reporter.bytes_written += cast(int, write_pipeline.buf_sz_bytes)
             else:
                 raise AssertionError(
                     "The completed task must be in either staging_tasks or io_tasks."
@@ -193,20 +321,14 @@ async def execute_write_reqs(
                 memory_budget_bytes=memory_budget_bytes,
                 executor=executor,
             )
-
-        logger.debug(
-            f"Rank {rank}\t"
-            f"ready_for_staging: {len(ready_for_staging)}\t"
-            f"staging_tasks: {len(staging_tasks)}\t"
-            f"ready_for_io: {len(ready_for_io)}\t"
-            f"io_tasks: {len(io_tasks)}\t"
-            f"memory_budget_bytes: {memory_budget_bytes}"
-        )
-
-    mbps = (bytes_written / 1e6) / (time.monotonic() - begin_ts)
-    logger.info(f"Rank {rank} finished saving. Throughput: {mbps:.2f}MB/s")
-
+    write_reporter.report_staging_done()
     executor.shutdown()
+    return PendingIOWork(
+        ready_for_io=ready_for_io,
+        io_tasks=io_tasks,
+        memory_budget_bytes=memory_budget_bytes,
+        write_reporter=write_reporter,
+    )
 
 
 def sync_execute_write_reqs(
@@ -215,10 +337,10 @@ def sync_execute_write_reqs(
     memory_budget_bytes: int,
     rank: int,
     event_loop: Optional[asyncio.AbstractEventLoop] = None,
-) -> None:
+) -> PendingIOWork:
     if event_loop is None:
         event_loop = asyncio.new_event_loop()
-    event_loop.run_until_complete(
+    return event_loop.run_until_complete(
         execute_write_reqs(
             write_reqs=write_reqs,
             storage=storage,
@@ -239,9 +361,9 @@ class _ReadPipeline:
         self.buf_sz_bytes: Optional[int] = None
 
     async def read_buffer(self) -> "_ReadPipeline":
-        io_req = IOReq(path=self.read_req.path)
-        await self.storage.read(io_req=io_req)
-        self.buf = io_req.buf.getvalue()
+        read_io = ReadIO(path=self.read_req.path, byte_range=self.read_req.byte_range)
+        await self.storage.read(read_io=read_io)
+        self.buf = read_io.buf.getvalue()
         self.buf_sz_bytes = len(self.buf)
         return self
 
@@ -281,7 +403,7 @@ async def execute_read_reqs(
                 len(io_tasks) == 0
                 or read_pipeline.consuming_cost_bytes < memory_budget_bytes
             ):
-                memory_budget_bytes += read_pipeline.consuming_cost_bytes
+                memory_budget_bytes -= read_pipeline.consuming_cost_bytes
                 io_task = asyncio.create_task(read_pipeline.read_buffer())
                 io_tasks.add(io_task)
                 dispatched_ids.add(i)
@@ -309,10 +431,10 @@ async def execute_read_reqs(
             if d in consuming_tasks:
                 consuming_tasks.remove(d)
                 read_pipeline: _ReadPipeline = d.result()
-                memory_budget_bytes -= read_pipeline.consuming_cost_bytes
+                memory_budget_bytes += read_pipeline.consuming_cost_bytes
                 bytes_read += cast(int, read_pipeline.buf_sz_bytes)
 
-    mbps = (bytes_read / 1e6) / (time.monotonic() - begin_ts)
+    mbps = (bytes_read / 1024**2) / (time.monotonic() - begin_ts)
     logger.info(f"Rank {rank} finished loading. Throughput: {mbps:.2f}MB/s")
 
     executor.shutdown()
